/**
   This code is released under the terms of the MIT license
   (https://opensource.org/licenses/MIT).

   Copyright 2019-2020 Stephan Beal (https://wanderinghorse.net)

   Permission is hereby granted, free of charge, to any person
   obtaining a copy of this software and associated documentation
   files (the "Software"), to deal in the Software without
   restriction, including without limitation the rights to use, copy,
   modify, merge, publish, distribute, sublicense, and/or sell copies
   of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be
   included in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
   NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
   BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
   ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
*/
/**
   This file contains the library-level code for the Dark Venture Card
   Tools. It is largely derived from earlier works by the same author
   for projects of a similar scope/intent. Thus it may (or may not)
   still contain references to and/or cruft regarding those projects.
*/
(function(global){
    

    // We explicitly don't support String objects - nobody uses those.
    const isString = (v)=>'string' === typeof v;
    const isArray = (v)=>v instanceof Array;
    const argsToArray = (a)=>Array.prototype.slice.call(a,0);
    const toss = function(msg/*...console args...*/){
        const a = argsToArray(arguments);
        const e = (msg instanceof Error ? msg : new Error(msg));
        a.shift();
        a.unshift("toss()ing Exception: ", e);
        console.error.apply(console,a);
        throw e;
    };
    const affirm = function(cond, msg){
        if(!cond) toss("Affirmation failed: "+(msg || "no further information"));
    };
    const affirmString = function(v,msg){
        if(isString(v)) return;
        toss(msg||"Expecting a string value.");
    };
    const affirmArray = function(v,msg){
        if(isArray(v)) return;
        toss(msg||"Expecting Array value.");
    };

    global.DarkVenture = {};
   
    /* DOM module */
    global.DarkVenture.dom = (function(){
        "use strict";
        /**
           A collection of HTML DOM utilities to simplify, a bit, using the
           DOM API. It is focused on manipulation of the DOM, but one of its
           core mantras is "No innerHTML." Using innerHTML in this code, in
           particular assigning to it, is absolutely verboten.
        */
        const argsToArray = (a)=>Array.prototype.slice.call(a,0);
        const isArray = (v)=>v instanceof Array;

        const dom = {
            create: function(elemType){
                return document.createElement(elemType);
            },
            createElemFactory: function(eType){
                return function(){
                    return document.createElement(eType);
                };
            },
            remove: function(e){
                if(e.forEach){
                    e.forEach(
                        (x)=>x.parentNode.removeChild(x)
                    );
                }else{
                    e.parentNode.removeChild(e);
                }
                return e;
            },
            /**
               Removes all child DOM elements from the given element
               and returns that element.

               If e has a forEach method (is an array or DOM element
               collection), this function instead clears each element in the
               collection. May be passed any number of arguments, each of
               which must be a DOM element or a container of DOM elements with
               a forEach() method. Returns its first argument.
            */
            clearElement: function f(e){
                if(!f.each){
                    f.each = function(e){
                        if(e.forEach){
                            e.forEach((x)=>f(x));
                            return e;
                        }
                        while(e.firstChild) e.removeChild(e.firstChild);
                    };
                }
                argsToArray(arguments).forEach(f.each);
                return arguments[0];
            },

            /**
               Evaluates whether e in the current viewport. Kinda.

               If the 2nd argument is falsy (the default), true is returned
               only if e is completely in the viewport (beware that elements
               larger than the viewport will always return false).

               If the 2nd argument is non-numeric truthy then it returns true
               if at least 1 of the top or bottom edge of e is in view. If the
               2nd argument is a number, true is returned if at least that
               many of the top/bottom edges are in view. (Note that we don't
               care whether the right/left edges are in view, as they're not
               relevant for what we do.)

               As a special case, if e's "display" style is set to
               "none", this routine always returns false.
            */
            isElementKindaInViewport: function f(e,partial) {
                // derived from https://stackoverflow.com/questions/123999/how-to-tell-if-a-dom-element-is-visible-in-the-current-viewport
                if("none"===window.getComputedStyle(e).display) return false;
                const rect = e.getBoundingClientRect(),
                      wh = (window.innerHeight || document.documentElement.clientHeight),
                      ww = (window.innerWidth || document.documentElement.clientWidth);
                
                if(partial){
                    let count = 0;
                    if(rect.top >= 0 && rect.top < wh) ++count;
                    if(rect.bottom <= wh && rect.bottom > 0) ++count;
                    //if(rect.left >= 0 && rect.left < ww) ++count;
                    //if(rect.right <= ww && rect.right > 0) ++count;
                    return ('number'===typeof partial) ? count >= partial : count > 0;
                }else{
                    // Return true only if it's completely visible
                    return (
                        rect.top >= 0 &&
                            rect.left >= 0 &&
                            rect.bottom <= (window.innerHeight || document.documentElement.clientHeight) &&
                            rect.right <= (window.innerWidth || document.documentElement.clientWidth)
                    );
                }
            },
            
            /**
               Tries to scroll the given element into its parent element's
               view. It is assumed that the element is a child of a parent
               in which it might be scrolled out of view (via
               overflow). (Note that a simple child.scrollIntoView() behaves
               horribly, scrolling the whole page.)

               Returns its argument.
            */
            scrollChildIntoView: function f(child){
                // Derived from https://stackoverflow.com/a/51005649
                if(!f.getRelativePos){
                    f.getRelativePos = function(e){
                        const pPos = e.parentNode.getBoundingClientRect(), // parent pos
                              cPos = e.getBoundingClientRect(), // target pos
                              pos = {};
                        pos.top    = cPos.top    - pPos.top + e.parentNode.scrollTop,
                        pos.right  = cPos.right  - pPos.right,
                        pos.bottom = cPos.bottom - pPos.bottom,
                        pos.left   = cPos.left   - pPos.left;

                        return pos;
                    };
                }
                const parent = child.parentNode;
                if(!parent) return child;
                const pos = f.getRelativePos(child);
                const start = parent.scrollTop,
                      change = pos.top - start;
                parent.scrollTop = start + change;
                return child;
            }
        }/* dom object */;

        /**
           Returns the result of splitting the given str on
           a run of spaces of (\s*,\s*).
        */
        dom.splitClassList = function f(str){
            if(!f.rx){
                f.rx = /(\s+|\s*,\s*)/;
            }
            return str ? str.split(f.rx) : [str];
        };
        
        dom.div = dom.createElemFactory('div');
        dom.p = dom.createElemFactory('p');
        dom.code = dom.createElemFactory('code');
        dom.pre = dom.createElemFactory('pre');
        dom.header = dom.createElemFactory('header');
        dom.footer = dom.createElemFactory('footer');
        dom.section = dom.createElemFactory('section');
        dom.span = dom.createElemFactory('span');
        dom.strong = dom.createElemFactory('strong');
        dom.em = dom.createElemFactory('em');
        /**
           Returns a LABEL element. If passed an argument,
           it must be an id or an HTMLElement with an id,
           and that id is set as the 'for' attribute of the
           label. If passed 2 arguments, the 2nd is text or
           a DOM element to append to the label.
        */
        dom.label = function(forElem, text){
            const rc = document.createElement('label');
            if(forElem){
                if(forElem instanceof HTMLElement){
                    forElem = this.attr(forElem, 'id');
                }
                dom.attr(rc, 'for', forElem);
            }
            if(text) this.append(rc, text);
            return rc;
        };
        /**
           Returns an IMG element with an optional src
           attribute value.
        */
        dom.img = function(src){
            const e = this.create('img');
            if(src) e.setAttribute('src',src);
            return e;
        };
        /**
           Creates and returns a new anchor element with the given
           optional href and label. If label===true then href is used
           as the label.
        */
        dom.a = function(href,label){
            const e = this.create('a');
            if(href) e.setAttribute('href',href);
            if(label) e.appendChild(dom.text(true===label ? href : label));
            return e;
        };
        dom.hr = dom.createElemFactory('hr');
        dom.br = dom.createElemFactory('br');
        dom.text = (t)=>document.createTextNode(t||'');
        dom.button = function(label){
            const b = this.create('button');
            if(label) b.appendChild(this.text(label));
            return b;
        };
        /**
           Returns a TEXTAREA element.

           Usages:

           ([boolean readonly = false])
           (non-boolean rows[,cols[,readonly=false]])

           Each of the rows/cols/readonly attributes is only set if it is
           truthy.
        */
        dom.textarea = function(){
            const rc = this.create('textarea');
            let rows, cols, readonly;
            if(1===arguments.length){
                if('boolean'===typeof arguments[0]){
                    readonly = !!arguments[0];
                }else{
                    rows = arguments[0];
                }
            }else if(arguments.length){
                rows = arguments[0];
                cols = arguments[1];
                readonly = arguments[2];
            }
            if(rows) rc.setAttribute('rows',rows);
            if(cols) rc.setAttribute('cols', cols);
            if(readonly) rc.setAttribute('readonly', true);
            return rc;
        };

        /**
           Returns a new SELECT element.
        */
        dom.select = dom.createElemFactory('select');

        /**
           Returns an OPTION element with the given value and label text
           (which defaults to the value).

           Usage:

           (value[, label])
           (selectElement [,value [,label = value]])

           Any forms taking a SELECT element append the new element to the
           given SELECT element.

           If any label is falsy and the value is not then the value is used
           as the label. A non-falsy label value may have any type suitable
           for passing as the 2nd argument to dom.append().

           If the value has the undefined value then it is NOT assigned as
           the option element's value and no label is set unless it has a
           non-undefined value.
        */
        dom.option = function(value,label){
            const a = arguments;
            var sel;
            if(1==a.length){
                if(a[0] instanceof HTMLElement){
                    sel = a[0];
                }else{
                    value = a[0];
                }
            }else if(2==a.length){
                if(a[0] instanceof HTMLElement){
                    sel = a[0];
                    value = a[1];
                }else{
                    value = a[0];
                    label = a[1];
                }
            }
            else if(3===a.length){
                sel = a[0];
                value = a[1];
                label = a[2];
            }
            const o = this.create('option');
            if(undefined !== value){
                o.value = value;
                this.append(o, this.text(label || value));
            }else if(undefined !== label){
                this.append(o, label);
            }
            if(sel) this.append(sel, o);
            return o;
        };
        dom.h = function(level){
            return this.create('h'+level);
        };
        dom.ul = dom.createElemFactory('ul');
        /**
           Creates and returns a new LI element, appending it to the
           given parent argument if it is provided.
        */
        dom.li = function(parent){
            const li = this.create('li');
            if(parent) parent.appendChild(li);
            return li;
        };

        /**
           Returns a function which creates a new DOM element of the
           given type and accepts an optional parent DOM element
           argument. If the function's argument is truthy, the new
           child element is appended to the given parent element.
           Returns the new child element.
        */
        dom.createElemFactoryWithOptionalParent = function(childType){
            return function(parent){
                const e = this.create(childType);
                if(parent) parent.appendChild(e);
                return e;
            };
        };
        
        dom.table = dom.createElemFactory('table');
        dom.thead = dom.createElemFactoryWithOptionalParent('thead');
        dom.tbody = dom.createElemFactoryWithOptionalParent('tbody');
        dom.tfoot = dom.createElemFactoryWithOptionalParent('tfoot');
        dom.tr = dom.createElemFactoryWithOptionalParent('tr');
        dom.td = dom.createElemFactoryWithOptionalParent('td');
        dom.th = dom.createElemFactoryWithOptionalParent('th');

        /**
           Creates and returns a FIELDSET element, optionaly with a LEGEND
           element added to it. If legendText is an HTMLElement then it is
           appended as-is, else it is assume (if truthy) to be a value
           suitable for passing to dom.append(aLegendElement,...).
        */
        dom.fieldset = function(legendText){
            const fs = this.create('fieldset');
            if(legendText){
                this.append(
                    fs,
                    (legendText instanceof HTMLElement)
                        ? legendText
                        : this.append(this.create('legend'),legendText)
                );
            }
            return fs;
        };

        /**
           Appends each argument after the first to the first argument
           (a DOM node) and returns the first argument.

           - If an argument is a string or number, it is transformed
           into a text node.

           - If an argument is an array or has a forEach member, this
           function appends each element in that list to the target
           by calling its forEach() method to pass it (recursively)
           to this function.

           - Else the argument assumed to be of a type legal
           to pass to parent.appendChild().
        */
        dom.append = function f(parent/*,...*/){
            const a = argsToArray(arguments);
            a.shift();
            for(let i in a) {
                var e = a[i];
                if(isArray(e) || e.forEach){
                    e.forEach((x)=>f.call(this, parent,x));
                    continue;
                }
                if('string'===typeof e
                   || 'number'===typeof e
                   || 'boolean'===typeof e) e = this.text(e);
                parent.appendChild(e);
            }
            return parent;
        };

        dom.input = function(type){
            return this.attr(this.create('input'), 'type', type);
        };
        /**
           Returns a new CHECKBOX input element.

           Usages:

           ([boolean checked = false])
           (non-boolean value [,boolean checked])
        */
        dom.checkbox = function(value, checked){
            const rc = this.input('checkbox');
            if(1===arguments.length && 'boolean'===typeof value){
                checked = !!value;
                value = undefined;
            }
            if(undefined !== value) rc.value = value;
            if(!!checked) rc.checked = true;
            return rc;
        };
        /**
           Returns a new RADIO input element.

           ([boolean checked = false])
           (string name [,boolean checked])
           (string name, non-boolean value [,boolean checked])
        */
        dom.radio = function(){
            const rc = this.input('radio');
            let name, value, checked;
            if(1===arguments.length && 'boolean'===typeof name){
                checked = arguments[0];
                name = value = undefined;
            }else if(2===arguments.length){
                name = arguments[0];
                if('boolean'===typeof arguments[1]){
                    checked = arguments[1];
                }else{
                    value = arguments[1];
                    checked = undefined;
                }
            }else if(arguments.length){
                name = arguments[0];
                value = arguments[1];
                checked = arguments[2];
            }
            if(name) this.attr(rc, 'name', name);
            if(undefined!==value) rc.value = value;
            if(!!checked) rc.checked = true;
            return rc;
        };

        /**
           Internal impl for addClass(), removeClass().
        */
        const domAddRemoveClass = function f(action,e){
            if(!f.rxSPlus){
                f.rxSPlus = /\s+/;
                f.applyAction = function(e,a,v){
                    if(!e || !v
                       /*silently skip empty strings/flasy
                         values, for user convenience*/) return;
                    else if(e.forEach){
                        e.forEach((E)=>E.classList[a](v));
                    }else{
                        e.classList[a](v);
                    }
                };
            }
            var i = 2, n = arguments.length;
            for( ; i < n; ++i ){
                let c = arguments[i];
                if(!c) continue;
                else if(isArray(c) ||
                        ('string'===typeof c
                         && c.indexOf(' ')>=0
                         && (c = c.split(f.rxSPlus)))
                        || c.forEach
                       ){
                    c.forEach((k)=>k ? f.applyAction(e, action, k) : false);
                    // ^^^ we could arguably call f(action,e,k) to recursively
                    // apply constructs like ['foo bar'] or [['foo'],['bar baz']].
                }else if(c){
                    f.applyAction(e, action, c);
                }
            }
            return e;
        };

        /**
           Adds one or more CSS classes to one or more DOM elements.

           The first argument is a target DOM element or a list type of such elements
           which has a forEach() method.  Each argument
           after the first may be a string or array of strings. Each
           string may contain spaces, in which case it is treated as a
           list of CSS classes.

           Returns e.
        */
        dom.addClass = function(e,c){
            const a = argsToArray(arguments);
            a.unshift('add');
            return domAddRemoveClass.apply(this, a);
        };
        /**
           The 'remove' counterpart of the addClass() method, taking
           the same arguments and returning the same thing.
        */
        dom.removeClass = function(e,c){
            const a = argsToArray(arguments);
            a.unshift('remove');
            return domAddRemoveClass.apply(this, a);
        };

        /**
           Toggles CSS class c on e (a single element for forEach-capable
           collection of elements. Returns its first argument.
        */
        dom.toggleClass = function f(e,c){
            if(e.forEach){
                e.forEach((x)=>x.classList.toggle(c));
            }else{
                e.classList.toggle(c);
            }
            return e;
        };

        /**
           Returns true if DOM element e contains CSS class c, else
           false.
        */
        dom.hasClass = function(e,c){
            return (e && e.classList) ? e.classList.contains(c) : false;
        };

        /**
           Each argument after the first may be a single DOM element or a
           container of them with a forEach() method. All such elements are
           appended, in the given order, to the dest element using
           dom.append(dest,theElement). Thus the 2nd and susequent arguments
           may be any type supported as the 2nd argument to that function.

           Returns dest.
        */
        dom.moveTo = function(dest,e){
            const n = arguments.length;
            var i = 1;
            const self = this;
            for( ; i < n; ++i ){
                e = arguments[i];
                this.append(dest, e);
            }
            return dest;
        };
        /**
           Each argument after the first may be a single DOM element
           or a container of them with a forEach() method. For each
           DOM element argument, all children of that DOM element
           are moved to dest (via appendChild()). For each list argument,
           each entry in the list is assumed to be a DOM element and is
           appended to dest.

           dest may be an Array, in which case each child is pushed
           into the array and removed from its current parent element.

           All children are appended in the given order.

           Returns dest.
        */
        dom.moveChildrenTo = function f(dest,e){
            if(!f.mv){
                f.mv = function(d,v){
                    if(d instanceof Array){
                        d.push(v);
                        if(v.parentNode) v.parentNode.removeChild(v);
                    }
                    else d.appendChild(v);
                };
            }
            const n = arguments.length;
            var i = 1;
            for( ; i < n; ++i ){
                e = arguments[i];
                if(!e){
                    console.warn("Achtung: dom.moveChildrenTo() passed a falsy value at argment",i,"of",
                                 arguments,arguments[i]);
                    continue;
                }
                if(e.forEach){
                    e.forEach((x)=>f.mv(dest, x));
                }else{
                    while(e.firstChild){
                        f.mv(dest, e.firstChild);
                    }
                }
            }
            return dest;
        };

        /**
           Adds each argument (DOM Elements) after the first to the
           DOM immediately before the first argument (in the order
           provided), then removes the first argument from the DOM.
           Returns void.

           If any argument beyond the first has a forEach method, that
           method is used to recursively insert the collection's
           contents before removing the first argument from the DOM.
        */
        dom.replaceNode = function f(old,nu){
            var i = 1, n = arguments.length;
            ++f.counter;
            try {
                for( ; i < n; ++i ){
                    const e = arguments[i];
                    if(e.forEach){
                        e.forEach((x)=>f.call(this,old,e));
                        continue;
                    }
                    old.parentNode.insertBefore(e, old);
                }
            }
            finally{
                --f.counter;
            }
            if(!f.counter){
                old.parentNode.removeChild(old);
            }
        };
        dom.replaceNode.counter = 0;        
        /**
           Two args == getter: (e,key), returns value

           Three or more == setter: (e,key,val[...,keyN,valN]), returns
           e. If val===null or val===undefined then the attribute is
           removed. If (e) has a forEach method then this routine is applied
           to each element of that collection via that method. Each pair of
           keys/values is applied to all elements designated by the first
           argument.
        */
        dom.attr = function f(e){
            if(2===arguments.length) return e.getAttribute(arguments[1]);
            const a = argsToArray(arguments);
            if(e.forEach){ /* Apply to all elements in the collection */
                e.forEach(function(x){
                    a[0] = x;
                    f.apply(f,a);
                });
                return e;
            }
            a.shift(/*element(s)*/);
            while(a.length){
                const key = a.shift(), val = a.shift();
                if(null===val || undefined===val){
                    e.removeAttribute(key);
                }else{
                    e.setAttribute(key,val);
                }
            }
            return e;
        };

        const enableDisable = function f(enable){
            var i = 1, n = arguments.length;
            for( ; i < n; ++i ){
                let e = arguments[i];
                if(e.forEach){
                    e.forEach((x)=>f(enable,x));
                }else{
                    e.disabled = !enable;
                }
            }
            return arguments[1];
        };

        /**
           Enables (by removing the "disabled" attribute) each element
           (HTML DOM element or a collection with a forEach method)
           and returns the first argument.
        */
        dom.enable = function(e){
            const args = argsToArray(arguments);
            args.unshift(true);
            return enableDisable.apply(this,args);
        };
        /**
           Disables (by setting the "disabled" attribute) each element
           (HTML DOM element or a collection with a forEach method)
           and returns the first argument.
        */
        dom.disable = function(e){
            const args = argsToArray(arguments);
            args.unshift(false);
            return enableDisable.apply(this,args);
        };

        
        /**
           Iterates over all child notes of DOM element e. For each
           TEXT_NODE-type node, it passes the node and e (in that
           order) to the given callback.

           This function iterates over the children in such a way that
           modifying e's child node list during iteration does not
           invalidate the iteration list.
           
           Returns e.
        */
        dom.visitTextNodes = function(e,callback){
            const ch = [];
            e.childNodes.forEach(function(c){
                if(c.nodeType===c.TEXT_NODE) ch.push(c);
            });
            // Because we may modify parent.childNodes while
            // iterating, we have to iterate over a copy of the
            // list...
            if(ch.length) ch.forEach((c)=>callback(c,e));
            return e;
        };

        /**
           If e is one of the following DOM element type(s):

           SPAN

           or if e.nodeType is not ELEMENT_NODE...

           then this is a no-op, else...

           Iterates over all child nodes of DOM element e. For each
           TEXT_NODE-type node, it replaces that node with a new node
           of type nodeType and appends that text node to that new
           node. nodeType must be a value suitable for use with
           document.createElement() and defaults to 'span' if no
           value, or a falsy value, is passed in.

           This function iterates over the children in such a way that
           modifying e's child node list during iteration does not
           invalidate the iteration list.

           If the 3rd argument is truthy, this transformation is applied
           recursively to e and its children, their children, etc. (if they
           are of the appropriate types).

           It can optionally be called in the form (node, boolean),
           which is equivalent to (node, false, boolean).

           Returns e.

           Background: the CSS'+' combinator selector does not consider text
           nodes when checking if its LHS and RHS are immediate siblings.
           That can lead to CSS which styles on (X + X) to behave in
           undesired ways when there are text nodes between two X
           elements. By wrapping the text nodes in non-text nodes, we can
           get consistent styling for such cases. A concrete example is:

           .foo + .foo { padding-left: 0.5em }

           <span class='foo'></span> text node </span class='foo'></span>

           As written, that will add left padding to the rightmost .foo
           element. This function wraps the "text node" part into a non-text
           node, meaning that the CSS no longer matches and no extra padding
           is injected.
        */
        dom.wrapTextNodes = function f(e, nodeType, recursive){
            if(!f.map){
                const m = f.map = {
                    //DIV: true, SECTION:true // inclusive map
                    //SPAN: true, // exclusive map
                };
                m.SPAN = m.A = true;
            }
            if(e.nodeType !== e.ELEMENT_NODE
               || !e.childNodes.length
               || f.map[e.tagName]) return e;
            else if(arguments.length===2 && 'boolean'===typeof nodeType){
                recursive = nodeType;
                nodeType = false;
            }
            if(!nodeType) nodeType = 'span';
            const recurse = recursive ? [] : undefined;
            if(recurse){
                // To aovid endless recursion, we need to capture the
                // list of non-text nodes before we replace the text
                // nodes...
                e.childNodes.forEach(function(c){
                    if(c.TEXT_NODE!==c.nodeType) recurse.push(c);
                });
            }
            this.visitTextNodes(e, function(child,parent){
                const n = document.createElement(nodeType);
                parent.insertBefore(n, child);
                n.appendChild(child);
            });
            if(recurse && recurse.length){
                recurse.forEach((c)=>f.call(this, c, nodeType, true));
            }
            return e;
        };


        /**
           Removes child nodes of DOM element e which of element type
           nodeType. If the 3rd argument is truthy, this operates
           recursively on all children of e which are not of type nodeType.

           Examples:

           dom.removeNodesOfType(e, e.COMMENT_NODE, true);
           dom.removeNodesOfType(e, e.TEXT_NODE, false);

           Background: empty text nodes, introduced by whitespace separating
           "relevant" DOM elements in a document, can mess up flexbox
           layouts. CSS gives us no way to select/address text nodes (e.g.,
           in order to set them to display:none), so... just nuke them.

           Returns e.
        */
        dom.removeNodesOfType = function f(e,nodeType,recurse){
            const ch = [], sub = recurse ? [] : undefined;
            e.childNodes.forEach(function(c){
                if(c.nodeType===nodeType) ch.push(c);
                else if(sub) sub.push(c);
            });
            if(ch.length) ch.forEach((c)=>e.removeChild(c));
            if(sub && sub.length) sub.forEach((s)=>f(s,nodeType,true));
            return e;
        };
        
        /**
           A proxy for document.querySelector() which throws if
           selection x is not found. It may optionally be passed an
           "origin" object as its 2nd argument, which restricts the
           search to that branch of the tree.
        */
        dom.selectOne = function(x,origin){
            var src = origin || document,
                e = src.querySelector(x);
            if(!e){
                e = new Error("Cannot find DOM element: "+x);
                console.error(e, src);
                throw e;
            }
            return e;
        };

        /**
           "Blinks" the given element a single time for the given number of
           milliseconds, defaulting (if the 2nd argument is falsy or not a
           number) to flashOnce.defaultTimeMs. If a 3rd argument is passed
           in, it must be a function, and it gets callback back at the end
           of the asynchronous flashing processes.

           This will only activate once per element during that timeframe -
           further calls will become no-ops until the blink is
           completed. This routine adds a dataset member to the element for
           the duration of the blink, to allow it to block multiple blinks.

           If passed 2 arguments and the 2nd is a function, it behaves as if
           it were called as (arg1, undefined, arg2).

           Returns e, noting that the flash itself is asynchronous and may
           still be running, or not yet started, when this function returns.
        */
        dom.flashOnce = function f(e,howLongMs,afterFlashCallback){
            if(e.dataset.isBlinking){
                return;
            }
            if(2===arguments.length && 'function' ===typeof howLongMs){
                afterFlashCallback = howLongMs;
                howLongMs = f.defaultTimeMs;
            }
            if(!howLongMs || 'number'!==typeof howLongMs){
                howLongMs = f.defaultTimeMs;
            }
            e.dataset.isBlinking = true;
            const transition = e.style.transition;
            e.style.transition = "opacity "+howLongMs+"ms ease-in-out";
            const opacity = e.style.opacity;
            e.style.opacity = 0;
            setTimeout(function(){
                e.style.transition = transition;
                e.style.opacity = opacity;
                delete e.dataset.isBlinking;
                if(afterFlashCallback) afterFlashCallback();
            }, howLongMs);
            return e;
        };
        dom.flashOnce.defaultTimeMs = 400;
        /**
           A DOM event handler which simply passes event.target
           to dom.flashOnce().
        */
        dom.flashOnce.eventHandler = (event)=>dom.flashOnce(event.target)

        /**
           Attempts to copy the given text to the system clipboard. Returns
           true if it succeeds, else false.
        */
        dom.copyTextToClipboard = function(text){
            if( window.clipboardData && window.clipboardData.setData ){
                window.clipboardData.setData('Text',text);
                return true;
            }else{
                const x = document.createElement("textarea");
                x.style.position = 'fixed';
                x.value = text;
                document.body.appendChild(x);
                x.select();
                var rc;
                try{
                    document.execCommand('copy');
                    rc = true;
                }catch(err){
                    rc = false;
                }finally{
                    document.body.removeChild(x);
                }
                return rc;
            }
        };

        /**
           Copies all properties from the 2nd argument (a plain object) into
           the style member of the first argument (DOM element or a
           forEach-capable list of elements). If the 2nd argument is falsy
           or empty, this is a no-op.

           Returns its first argument.
        */
        dom.copyStyle = function f(e, style){
            if(e.forEach){
                e.forEach((x)=>f(x, style));
                return e;
            }
            if(style){
                let k;
                for(k in style){
                    if(style.hasOwnProperty(k)) e.style[k] = style[k];
                }
            }
            return e;
        };

        /**
           Parses a string as HTML.

           Usages:

           Array (htmlString)
           DOMElement (DOMElement target, htmlString)

           The first form parses the string as HTML and returns an Array of
           all elements parsed from it. If string is falsy then it returns
           an empty array.

           The second form parses the HTML string and appends all elements
           to the given target element using dom.append(), then returns the
           first argument.

           Caveats:

           - It expects a partial HTML document as input, not a full HTML
           document with a HEAD and BODY tags. Because of how DOMParser
           works, only children of the parsed document's (virtual) body are
           acknowledged by this routine.
        */
        dom.parseHtml = function(){
            let childs, string, tgt;
            if(1===arguments.length){
                string = arguments[0];
            }else if(2==arguments.length){
                tgt = arguments[0];
                string  = arguments[1];
            }
            if(string){
                const newNode = new DOMParser().parseFromString(string, 'text/html');
                childs = newNode.documentElement.querySelector('body');
                childs = childs ? Array.prototype.slice.call(childs.childNodes, 0) : [];
                /* ^^^ we need a clone of the list because reparenting them
                   modifies a NodeList they're in. */
            }else{
                childs = [];
            }
            return tgt ? this.moveTo(tgt, childs) : childs;
        };
        return dom;
    })();
    
    /**
       The text markup processor module.
    */
    global.DarkVenture.MarkerUpper = (function(DV/*==>window.DarkVenture object*/){
        "use strict";
        const dom = DV.dom;
        /**
           A basic text markup processor which does not use innerHTML. It
           instead works only with the DOM API (except for innerHTML, that
           is).

           If passed boolean true or another MarkerUpper instance then
           this instance gets flagged as being a sub-markup. It does
           not associate itself with the other instance, but does skip
           certain processing steps which must only be performed once
           on any given input and were presumably performed by the
           top-most instance (basic unescaping and such).
        */
        const MarkerUpper = function MarkerUpper(){
            this.reset();
            const arg = arguments[0];
            if(true===arg || arg instanceof MarkerUpper){
                this.isRecursing = true;
            }
        };

        /**
           Internal helper to MarkerUpper recursively on a fragment of
           a document (the bodies of block-style tags).
        */
        const recurseMarkup = function(content, bodyTagType){
            const m = new MarkerUpper(true);
            m.domTagType = bodyTagType;
            return m.markup(content);
        };

        /**
           If v is falsy, this is a no-op, else
           tgt.push(v) is called.
        */
        const pushNoFalsy = function(tgt, v){
            if(v) tgt.push(v);
        };

        /**
           Replaces all escape sequences in the this.content and sets
           this.top[0] to that value.  Currently only '\n' ==> ASCII
           10d.
        */
        const p00Escapes = function f(){
            if(!f.rx) f.rx = /\\n/g;
            affirmString(this.content);
            this.top = [this.content.replace(
                f.rx, '\n'
            )];
            return this;
        };

        /**
           Replaces all runs of:

           1) 3 "-" characters with HR elements.

           2) 3 "=" characters with HR elements with the CSS class 'thick'.
        */
        const p20Separators = function f(){
            if(!f.rx){
                f.rxMinus = /-{3}\r?\n?/;
                f.rxMinus.cssClass = '';
                /* if i use (\r?\n)? then we end up with duplicated HRs, and i've no clue why. */
                f.rxEquals = /={3}\r?\n?/;
                f.rxEquals.cssClass = 'thick';
            }
            affirmArray(this.top, "Expecting Array this.top.");
            const doIt = function(rx, dest, line){
                if(!isString(line) || line.length<3){
                    pushNoFalsy(dest, line);
                    return;
                }
                const t = line.split(rx);
                t.forEach((p,ndx)=>{
                    if(ndx>0){
                        let hr = dom.hr();
                        if(rx.cssClass) dom.addClass(hr, rx.cssClass);
                        dest.push(hr);
                    }
                    pushNoFalsy(dest, p);
                });
            };
            let buffer = [];
            this.top.forEach((x)=>doIt(f.rxMinus, buffer, x));
            this.top = buffer;
            buffer = [];
            this.top.forEach((x)=>doIt(f.rxEquals, buffer, x));
            this.top = buffer;
            return this;
        };

        /**
           Replaces references in the form {:XYZ:} (keyword+glyph) and
           {::XYZ::} (glyph only), noting that {:XYZ:} is equivalent
           to {::XYZ::} for glyphs which have no associated keywords
           (e.g. the dice-face glyphs). (That wasn't planned, it's
           just a side effect which would be more trouble than it's
           worth to make more pedantic.)
        */
        const p25Keywords = function f(){
            if(!f.rxInnerGlyph){
                /* Test whether the "XYZ" part of {:XYZ:}
                   matches ^:XYZ:$ (i.e. {::XYZ::}), with
                   an optional :colorname suffix. */
                f.rxInnerGlyph = /^:\w+(:[-\w]+)?:$/;

                /**

                   Passed the XYZ part of {:XYZ:} or {::XYZ::}, this fishes
                   out a self.config.keywords entry, taking into account that
                   XYZ might be KEYWORD:colorname. If such a match is found,
                   it returns an object which has the same structure as the
                   self.config.keywords, but:

                   1) If the keyword entry's ddicon is set, the result object
                   has a SPAN containing that icon entry, as produced by
                   ddIconFontEntries.getElement()

                   1) With its img property adjusted for the specified
                   color (replacing /black/ with /colorname/) and a color
                   property containing colorname.

                   2) Its label property will be either the original label
                   (if it's a string or falsy) or the label[language]
                   value of the original entry.
                */
                f.kwdEntry = function(self,key){
                    var kwd = self.config.keywords[key];
                    const m = kwd ? false : key.split(':',2);
                    if(!kwd){
                        key = m[0];
                        kwd = self.config.keywords[key];
                        if(!kwd){
                            self.toss("Unknown keyword/glyph key: "+key);
                        }
                    }
                    const color =  m
                          ? (m[1]||'')
                          : undefined
                    const rc = {
                        label: (!kwd.label || 'string'===typeof kwd.label)
                            ? kwd.label : kwd.label[self.config.language],
                        color: color
                    };
                    if(kwd.ddicon){
                        rc.ddicon = ddIconFontEntries.getElement(kwd.ddicon,color);
                    }else{
                        /**
                           Only set this if kwd.ddicon is not set, as creation of this IMG
                           tag triggers loading of the SVG from the server. That's particularly
                           noticeable in the editor apps when they render their markup overview
                           widgets.
                        */
                        rc.img = kwd.img
                            ? dom.img(self.config.libDirUri + (
                                color ? kwd.img.replace('/black/', '/'+color+'/') : kwd.img
                            ))
                            : undefined;
                    }
                    return rc;
                };

            }
            const out = [];
            const self = this;
            const push = (v)=>pushNoFalsy(out, v);

            this.top.forEach(function(p){
                var brPos;
                if(!isString(p)
                   || p.length<5 /* length of {:X:} */
                   || (brPos=p.indexOf('{:')) < 0
                  ){
                    push(p);
                    return;
                }
                //Reminders:
                // string.substr(start, length); start may be negative offset from end
                // string.substring(start, end); start may not be negative
                while(brPos>=0){
                    if(brPos) push(p.substr(0, brPos));
                    p = p.substr(brPos+2);
                    const clPos = p.indexOf(':}');
                    if(clPos<1){ /* shortest valid ==> {:1:} */
                        self.toss("Malformed {:...:} group: "+p);
                    }
                    var inner = p.substr(0,clPos);
                    var glyphOnly = f.rxInnerGlyph.test(inner);
                    //console.debug("inner =",inner,"glyphOnly =",glyphOnly);
                    if(glyphOnly) inner = inner.substr(1,inner.length-2) /* strip 2 extra : */;
                    if(!inner){
                        self.toss("Empty "+(glyphOnly ? "{::GLYPH::}" : "{:KEYWORD:}")+" construct.");
                    }
                    const kwd = f.kwdEntry(self,inner);
                    if(!kwd.label){
                        /* Special case to get consistent behaviour
                           with {:#:} and {::#::}.
                        */
                        glyphOnly = true;
                    }
                    //console.debug("kwd =",kwd);
                    p = p.substr(clPos+2);
                    // Inject keyword/glyph {:here:}
                    const eKwd = kwd.ddicon || kwd.img;
                    if(!eKwd){
                        self.toss("Missing image for keyword/glyph: "+inner);
                    }else if(kwd.img===eKwd){
                        dom.addClass(kwd.img, self.config.css.cardGlyph);
                    }
                    const e = glyphOnly ? undefined : dom.span();
                    if(e){
                        dom.addClass( e, self.config.css.keywordRef );
                        if(kwd.label){
                            const sp = dom.addClass(dom.span(),'keyword');
                            if(kwd.color){
                                dom.addClass(e, 'dv-fg-'+kwd.color);
                            }
                            e.appendChild(dom.append(sp,dom.text(kwd.label)));
                        }
                        e.appendChild(eKwd);
                        push(e);
                    }else{
                        push(eKwd);
                    }
                    brPos = p.indexOf('{:');
                }
                push(p);
            });
            //console.debug("out =",out);
            this.top = out;
            return this;
        };

        /**
           Replaces all character attribute modifier markup in the
           given text and returns the (potentially modified)
           string. Attribute modifier markup is: [[+n ATTR]] or [[-1
           ATTR]], e.g. [[+1 STR]]. This function does not enforce the
           order of the modifier and attribute, but it will "look
           wrong" if they're reversed in the input. It does enforce
           that each such block contains exactly two tokens, and it
           throws if that is violated.

           The generated markup has an outer DIV.dv-attr-mod, one
           inner DIV (and unfortunate requirement to allow us to mix
           display: inline-block with flex layout), and two SPANs in
           that inner DIV.
        */
        const p50AttrMod = function f(){
            const out = [];
            const self = this;
            this.top.forEach(function(p){
                var brPos;
                if(!isString(p)
                   || p.length<10
                   || (brPos=p.indexOf('[[')) < 0
                  ){
                    pushNoFalsy(out,p);
                    return;
                }
                //Reminders:
                // string.substr(start, length); start may be negative offset from end
                // string.substring(start, end); start may not be negative
                while(brPos>=0){
                    if(brPos>0) pushNoFalsy(out, p.substr(0, brPos));
                    p = p.substr(brPos+2);
                    const clPos = p.indexOf(']]');
                    if(clPos<=5 /* "+X YZ"*/){
                        self.toss("Malformed [[...]] group.");
                    }
                    const inner = p.substr(0,clPos);
                    p = p.substr(clPos+2);
                    // Inject attr modifier [[here]].
                    const halves = inner.split(/\s+/);
                    if(2 !== halves.length){
                        self.error("attr modifier halves =<<<",
                                   halves, ">>>< inner = (((",inner,")))");
                        self.toss("Malformed attribute modifier! Expecting: [[+n ATT]]");
                    }
                    const div = dom.div();
                    dom.addClass(
                        div,
                        self.config.css[
                            ('-'===halves[0][0]) ? 'attrModMinus' : 'attrModPlus'
                        ]
                    );
                    out.push(div);
                    const divKludge = dom.div();// browser layout kludge
                    div.appendChild(divKludge);
                    dom.append(divKludge,
                               dom.append(dom.span(),halves[0]),
                               // // no separator needed because CSS uses a flexbox to lay these out
                               dom.append(dom.span(),halves[1])
                              );
                    brPos = p.indexOf('[[');
                }
                pushNoFalsy(out,p);
            });
            this.top = out;
            return this;
        };


        /**
           Internal helper for certain markup tag handlers.

           Usages:

           1) inlineMarkup(tgt, txt) processes txt as markup,
           empties out tgt, moves all resulting child elements
           into tgt (a DOM element or array), then returns
           tgt.

           2) inlineMarkup(txt) processes txt as markup and returns
           the result from MarkerUpper.markup() (or an empty DIV if
           txt was empty).
        */
        const inlineMarkup = function f(tgt,txt){
            if(1===arguments.length){
                txt = arguments[0];
                tgt = undefined;
            }
            const m = new MarkerUpper();
            m.inlineOnly = true;
            //m.domTagType = 'span'; /* currently breaks card header layout */
            const marked = m.markup(txt)||dom.span();
            return tgt
                ? dom.moveChildrenTo(
                    tgt instanceof Array ? tgt : dom.clearElement(tgt),
                    marked
                ) : marked;
        };

        /**
           Internal helper for the TagHandlers bits. Expects tagArgs
           to be the XXX part of {{tag[=]XXX}}. If its a string, it
           gets trimmed and basic sanity checks are applied, throwing
           on error. If it's not a string, it is ingored (under the
           assumption that it may have already been processed as part
           of a recursive call). Returns the (possibly transformed)
           tagArgs.

           If tagHandler.tagArgsType is one of the following values,
           tagArgs is treated as described below:

           'strings-space' = gets split() on \s+ and returned as an
           array.

           'json' = if it's not empty, after trimming, it's passed to
           JSON.parse(), and any parsing exception is propagated via
           marker.toss(). It may be an Object or Array.
        */
        const preprocessTagArgs = function(marker, tagName, tagHandler, tagArgs){
            if('string'===typeof tagArgs){
                // ^^^ during recursive/proxied calls, this might be an array or object.
                tagArgs = tagArgs.trim();
                if('=' === tagArgs[0]){
                    tagArgs = tagArgs.substr(1);
                }else if(tagHandler.requiresEquals){
                    marker.toss("Missing '=' after tag "+tagName+".");
                }
                const tt = tagHandler.tagArgsType;
                if('strings-space'===tt){
                    tagArgs = marker.splitSpaceDelim(tagArgs);
                }else if('json'===tt){
                    if(tagArgs){
                        try{tagArgs = JSON.parse(tagArgs)}
                        catch(e){marker.toss(e,". JSON Input data:",tagArgs)}
                    }
                }
            }
            return tagArgs;
        };

        /**
           Internal helper for handling tag body text before passing it on to
           the enclosing tag handler.
        */
        const processTagBody = function(marker, tagName, tagHandler, body){
            if(tagHandler.processBody && 'string'===typeof body){
                // ^^^ during recursive/proxied calls, this might be an array or object.
                const pb = tagHandler.processBody;
                if(true===pb){
                    return recurseMarkup(body, tagHandler.bodyTagType);
                }else if('json'===pb){
                    if(body){
                        try{return JSON.parse(body)}
                        catch(e){marker.toss("Error parsing JSON tag body:",body, e)}
                    }else{
                        marker.toss("Empty JSON body is not permitted (tag="+tagName+")");
                    }
                }
                marker.toss("Unhandled TagHandlers."+tagName+".processBody value: ["+pb+"]");
            }
            return body;
        };

        /**
           This routine is a helper for TagHandler implementations
           which accept optional [bg [, fg [,css classes...]]]
           arguments.

           Requires marker to be a MarkerUpper instance, tgtElem to be
           a DOM element (the one passed to the TagHandler's process()
           method), and args to be an array (the arguments passed to
           process()). The 4th argument is an optional object with
           default config options (see below).

           This routine consumes all of the array, applying some
           framework-internal-conventional logic:

           - If the first value looks like a color, it is set as a
           dv-bg-xxx value.

           - If the second value looks like a color, it is set as a
           dv-fg-xxx value.

           - If any other value is a legal font hint size integer, it
           is set as a dv-font-size-xx class.

           - Any other value is assumed to be a CSS class name.

           If the first two arguments are NOT color names then opt.bg
           resp. opt.fg are checked, and those values are used if they are
           set.

           All such classes are applied to tgtElem.

           The given array is consumed by this operation and will, on
           success, be empty upon returning.

           Obscure opt option(s):

           - swapFgBg: if truthy, this function swaps the order of the expected
           bg/fg arguments, interpreting them as fg/bg instead..

           Returns tgtElem.
        */
        const pushExtraTagStyleArgs = function(marker,tgtElem,args,opt){
            //console.debug("args =",args,"opt =",opt);
            var ndx = 0, gotBg, gotFg;
            for( ; args.length; ++ndx){
                const arg = args.shift();
                //console.debug(ndx, arg);
                if(ndx===(opt.swapFgBg ? 1 : 0)){
                    if(arg && marker.isColorName(arg)){
                        gotBg = arg;
                        marker.setBgClass(tgtElem,arg);
                        continue;
                    }
                    /* Fall through */
                }
                if(ndx===(opt.swapFgBg ? 0 : 1)){
                    if(arg && marker.isColorName(arg)){
                        gotFg = arg;
                        marker.setColorClass(tgtElem,arg);
                        continue;
                    }
                    /* Fall through */
                }
                if(marker.isLegalFontSizeHint(+arg)) marker.setFontSizeHint(tgtElem,+arg);
                else if(arg) dom.addClass(tgtElem,arg);
            }
            if(opt && !gotBg && opt.bg) marker.setBgClass(tgtElem, opt.bg);
            if(opt && !gotFg && opt.fg) marker.setColorClass(tgtElem, opt.fg);
            return tgtElem;
        };

        /**
           Internal helper for blocks in the forms:

           {{key[[=]data...]}}...{{/key}}

           {{key[[=]data...]}}
        */
        const TagHandlers = {
            /**
               The API for a markup tag handler object...
            */
            _api_doc_placeholder:{
                /**
                   If a block handler's requireEquals is true the this
                   API pedantically requires a '=' directly after the
                   key, otherwise it does not. This is a matter of
                   being intentionally strict, possibly too much so,
                   which can possibly be changed later.

                   20190822: this is no longer preferred. Readability
                   is improved by not having a '='. It is retained for
                   compatibility, but no tags require an equal sign any
                   more.
                */
                requiresEquals: false,

                /**
                   The DOM element type to be used for representing
                   this handler's top-most element. This defaults to
                   DIV, but may need to be set to SPAN for certain
                   types of elements.
                */
                bodyTagType: undefined,

                /**
                   A hint to the downstream processor as to whether
                   this markup is considered "block" or "inline", the
                   difference being one of finely-split
                   hairs. "inline" handlers are expected to output a
                   non-block DOM element, e.g. a SPAN or a DIV with a
                   CSS class which sets its display to inline-block
                   (or similar, including a list of such elements).
                   Block markup is expected to resolve to block-level
                   elements.

                   "Inline" markup is allowed in some contexts and not
                   others, and non-inline markup will trigger an
                   exception when it's not allowed.

                   If a handler does not set this property,
                   getHandler() sets to to true by default. If the
                   property sets isInline to truthy then getHandler()
                   sets this property to false.
                */
                isBlock: true,

                /**
                   If hasClosingTag is truthy (the default) then the
                   markup is assumed to be in the form
                   {{tag...}}...{{/tag}}, else it's assumed to be in
                   the form {{tag ...data...}}, with no closing
                   {{/tag}}.

                   For the no-closer case the markup processor passes
                   the undefined value as the 4th argument to its
                   process() method and expects process() to return a
                   DOM element (or list of elements) with which to
                   replace the {{tag}}.
                */
                hasClosingTag: true,

                /**
                   EXPERIMENTAL.

                   A hint to the processor that it is allowed
                   to/should trim the inner body of block-level markup
                   before recursing into its contents. Oftentimes
                   trimming can cause unwanted layout effects, so
                   should only be used when it's known to be okay.

                   If not falsy, this may be any of'trim', 'trimLeft',
                   or 'trimRight', and that method of the String class
                   will be used to perform the
                   trim.
                */
                trimBodyMethod: false,

                /**
                   If falsy, the processor will not recurse into a
                   tag's body content (if any), and will instead pass
                   it on as-is (string, possibly empty) to the tag
                   handler.

                   If boolean true (the default) the processor will
                   recurse into the tag's body content.

                   If it's "json" then the processor will parse the
                   body content as JSON and pass that object to the
                   tag handler.

                   Any unknown/unhandled value will trigger an
                   exception.
                */
                processBody: true,

                /**
                   An optional string describing how the XXX part of {{tag[=]?XXX}}
                   should be preprocessed before calling this TagHandler's process
                   method. See preprocessTagArgs() for the various
                   legal values. Unknown values are ignored.
                */
                tagArgsType: undefined,

                /**
                   This method is responsible for processing the
                   markup tag. When called by this framework, this tag
                   handler object will be the current "this".

                   Args:

                   1) MarkerUpper instance. The primary intention is
                   that its toss() method be used to report errors. It
                   can also be used for higher-level utility methods.

                   2) The XXX part of {{XXX[=...]}}

                   3) The YYY... part of {{XXX[=]YYY...}}

                   4) The DOM element which holds the body which this
                   function should wrap in its own block-level
                   element.  If this object's hasClosingTag is falsy
                   then this will be a falsy value.

                   The processor must:

                   1) Call marker.toss() to report errors.

                   2) Return either a DOM element or a list of DOM
                   elements (any list type with a forEach() method)
                   with which the calling processor will replace the
                   tag(s). If the markup processor's hasClosingTag is
                   truthy then it is assumed that the process method
                   will either enclose the passed-in bodyElem in new
                   tags or otherwise consume the bodyElem data.
                   e.g. the {{colorbox}} markup wraps bodyElem in a
                   DIV which contains the colorbox's style-related
                   properties, then returns that DIV.
                */
                process:function( marker, tag, tagArgs, bodyElem ){
                    marker.toss("Don't call this one!")
                }
            },

            /**
               {{.}} is replaced with a span wrapping a \u2219 - a
               center dot used as a field separator for card data in
               various contexts. The span has hte class
               'bullet-operator' (the Unicode name of that character)
               so that its size can be increased for the miniscule default.
            */
            ".":{
                isBlock: false,
                hasClosingTag: false,
                process: function(marker, tag, tagArgs, bodyElem){
                    return dom.addClass(
                        dom.append(dom.span(), "∙"),
                        'bullet-operator'
                    );
                }
            },

            /**
               Wraps its body element in a div with the classes
               'align-text' and 'X', where X is the value passed to
               this block via {{align=X}}.
            */
            align:{
                isBlock: true,
                process:function( marker, tag, tagArgs, bodyElem ){
                    const div = dom.div();
                    dom.addClass(div, "align-text", tagArgs);
                    div.appendChild(bodyElem);
                    return div;
                }
            },

            br:{
                isBlock: false,
                hasClosingTag: false,
                process: function(marker, tag, tagArgs, bodyElem){
                    return dom.br();
                }
            },

            /**
               Handler for:

               {{colorbox[=]bgcolor [fgcolor [css classes...]]}}...{{/colorbox}}.

               If bodyElem is truthy, it gets decorated with classes
               dv-bg-{bgcolor} and (if provided) dv-fg-{fgcolor}.
               Any additional arguments are assumed to be CSS classes
               and set as-is on the element.

               If bodyElem is falsy, a new DIV is created.

               Returns bodyElem resp. its new DIV.
            */
            colorbox:{
                isBlock: true,
                trimBodyMethod: 'trim',
                tagArgsType: 'strings-space',
                process: function(marker, tag, tagArgs, bodyElem){
                    const div = bodyElem || dom.div();
                    dom.addClass(div, "dv-colorbox");
                    pushExtraTagStyleArgs(marker, div, tagArgs, {bg:"default",fg:"default"});
                    return div;
                }
            },

            
            /**
               Replaces instances of {{fontsize=xxx}}...{{/fontsize}}
               with a DIV which has a hard-coded "style" attribute of:
               font-size:xxx. This is generally considered to be a
               last resort for configuring font sizes, but it can
               reasonably be used to set relative font sizes, so long
               as the size difference is not enough to throw the
               overall layout out of whack. The font size part must
               not contain any double-quotes!
            */
            fontsize:{
                isBlock: true,
                process: function(marker, tag, tagArgs, bodyElem){
                    const div = dom.div();
                    dom.attr(div, "style", "font-size:"+tagArgs);
                    div.appendChild(bodyElem);
                    return div;
                }
            },

            /**
               {{glyph [color/]basename [optional CSS classes]}}

               Resolves to an IMG element with
               src=marker.getGlyphDirUri(color/basename.svg), where
               the color defaults to black if the argument contains no
               '/'.

               If tagArgs contains any tokens after [color/]basename,
               each one (space-delimited) is set as a CSS class on the
               IMG tag, else the IMG tag gets the single CSS class
               marker.config.css.cardGlyph added to it.
            */
            glyph:{
                isBlock: false,
                hasClosingTag: false,
                tagArgsType: 'strings-space',
                process: function(marker, tag, tagArgs, bodyElem){
                    affirm(!bodyElem,tag+" markup expects no body element!");
                    //console.debug('glyph arguments =',arguments);
                    const tags = tagArgs;
                    if(!tags || !tags.length){
                        marker.toss("Missing color/name in {{"+tag+"}} arguments.");
                    }
                    const src = marker.getGlyphUri(tags.shift());
                    const img = dom.img(src);
                    if(tags.length) dom.addClass(img, tags);
                    else dom.addClass(img, marker.config.css.cardGlyph);
                    return img;
                }
            },


            /**
               {{nowrap ...}}...{{/nowrap}} is shorthand for:

               {{span nowrap ...}}...{{/span}}
            */
            nowrap:{
                isInline: true,
                tagArgsType: 'strings-space',
                process: function f(marker, tag, args, bodyElem){
                    return dom.moveChildrenTo(
                        dom.addClass(dom.span(), 'dv-nowrap', args),
                        bodyElem
                    );
                }
            },

            /**
               {{span css-classes...}}...{{/span}}
            */
            span:{
                isInline: true,
                tagArgsType: 'strings-space',
                process: function f(marker, tag, args, bodyElem){
                    if(!bodyElem){
                        if(!args.length) return dom.span();
                        bodyElem = dom.text('');
                    }
                    else if('SPAN'===bodyElem.tagName && !bodyElem.className){
                        /*
                          Rather than create nested spans, when the inner
                          one has no CSS classes of its own, we'll simply
                          re-use it as our result, adding our classes (if any)
                          to it...
                        */
                        dom.addClass(bodyElem, args);
                        return bodyElem;
                    }
                    if(!f.take){
                        const m = f.take = {};
                        m.DIV = m.P = true;
                    }
                    const rc = dom.addClass(dom.span(), args);
                    if(bodyElem.nodeType === bodyElem.ELEMENT_NODE){
                        // Kludgy/fragile implementation, but it "should"
                        // suffice for our purposes...
                        if(f.take[bodyElem.tagName]){
                            dom.moveChildrenTo(rc, bodyElem);
                        }else{
                            dom.append(rc, bodyElem);
                        }
                    }else if(bodyElem.nodeType === bodyElem.TEXT_NODE){
                        dom.append(rc, bodyElem);
                    }else{
                        console.warn("{{span}} doesn't know how to handle body:",bodyElem);
                    }
                    return rc;
                }
            },

            /**
               Resolves to an asterisk wrapped in a text node (wrapped
               in an enigma...). This is the workaround for injecting
               a literal asterisk, since we otherwise use asterisks as
               markup with markdown-like semantics.
            */
            asterisk:{
                isInline: true,
                hasClosingTag: false,
                process: function f(marker, tag, nil, nil2){
                    return dom.text('*');
                }
            },
            
            /**
               Huge kludge:

               {{url -s the-url ...}}

               which is almost equivalent to:

               {{span word-wrap-break-word ...}}the-url{{/span}}

               Except that this handler also hyperlinks the URL and
               sets the CSS class style-like-text so that CSS can
               remove the conventional underlining and have the link
               inherit its colors from the parent.
            */
            url:{
                isInline: true,
                tagArgsType: 'strings-space',
                hasClosingTag: false,
                process: function f(marker, tag, args, nil){
                    var arg = args.shift();
                    if('-s' === arg){
                        arg = args.shift();
                        if(arg) arg = arg.replace(/^\w+:\/\//,'');
                    }
                    const usrc = arg ||'MISSING {{url}} argument';
                    args.unshift('style-like-text');
                    const url = dom.addClass(
                        dom.a(
                            usrc,
                            true
                        ),
                        args
                    );
                    //url.style.color = url.style.backgroundColor = 'inherit';
                    //url.style.textDecoration = 'none';
                    //const rc = TagHandlers.span.process(
                    //    marker, 'span', [], url
                    //);
                    return url;
                }
            },
            /**
               Huge kludge which is equivalent to:

               {{span word-wrap-break-word ...other args}}...{{/span}}
            */
            wrapword:{
                isInline: true,
                tagArgsType: 'strings-space',
                process: function f(marker, tag, args, bodyElem){
                    args.unshift('word-wrap-break-word');
                    return TagHandlers.span.process(marker, 'span', args, bodyElem);
                }
            },
            
            /**
               Returns the tag handler for the given tag name, or a
               falsy value if no such handler is found. It may modify
               the handler before returning it, setting or normalizing
               default property values.
            */
            getHandler: function(tag){
                const h = (
                    this.hasOwnProperty(tag)
                        ? ((this[tag] && this[tag].process) ? this[tag] : null)
                        : null
                );
                if(h){
                    // Finagle some of the BlockHandler settings, setting defaults and normalizing them...
                    if(!h.hasOwnProperty('hasClosingTag')) h.hasClosingTag = true;
                    if(!h.hasOwnProperty('processBody')) h.processBody = true;
                    if(!h.hasOwnProperty('isBlock')) h.isBlock = !h.isInline;
                    if(!h.hasOwnProperty('isInline')) h.isInline = !h.isBlock;
                    if((h.isInline && h.isBlock) || (!h.isInline && !h.isBlock)){
                        toss("TagHandler misconfiguration ("
                             +tag+"): isInline and isBlock are both "
                             +(!!h.isInline)+".");
                    }
                    h.isInline = !h.isBlock;
                }                
                return h;
            },

            /**
               Uses getHandler(tag) to find a processor function and
               calls processor.process(marker, tag, tagArgs (possibly
               modified), bodyElem).  Throws on error. This function
               strips, if found, any one leading '=' from tagArgs.
               If the requiresEquals property of the handler is true
               then it affirms that tagArgs starts with '=', throwing
               if that's not the case.

               All exceptions are thrown by calling marker.toss().
            */
            processTag: function(marker, tag, tagArgs, bodyElem){
                const h = this.getHandler(tag);
                if(!h || 'object'!==typeof h){
                    marker.toss("No handler found for tag '"+tag+"'.");
                }
                tagArgs = preprocessTagArgs(marker, tag, h, tagArgs);
                return h.process(marker, tag, tagArgs, bodyElem);
            }
        }/*TagHandlers*/;

        /**
           TagHandlers impl for the {{b}}, {{i}}, and {{bi}} tags.
        */
        const _processBIBI = function f(marker, tag, ignored, bodyElem){
            if(!f.map){
                f.map = {
                    // tag==>CSS class(es).
                    b: 'dv-strong',
                    i: 'dv-italic',
                    bi: ['dv-strong','dv-italic'],
                    u: ['dv-underline'],
                    lowercase: 'dv-lowercase'
                };
            }
            //console.debug("BIBI "+tag+" body =",bodyElem);
            const span = 0 ? dom.span() : bodyElem;
            const cssClass = f.map[tag];
            if(!cssClass){
                marker.toss(
                    "Internal error: missing CSS class mapping for tag "+tag+"."
                );
            }
            dom.addClass(span, cssClass);
            if(span !== bodyElem){
                dom.moveChildrenTo(
                    span, bodyElem
                );
            }
            return span;
        };

        ['b','i','bi', 'u', 'lowercase'
         /* {{lowercase}} is an undocumented internal-only markup
            kludge to handle the lower-cased (by convention) "or" in
            card requirement lists. (20190813: is it still used?) */
        ].forEach((t)=>TagHandlers[t] = {
            requiresEquals: false,
            isInline: true,
            bodyTagType: 'span' /*tells recurseMarkup() to use a SPAN
                                  instead of a DIV for the body*/,
            process: _processBIBI
        });
        
        /**
           Transform certain markdown-style tags to {{...}} tags to
           avoid having to reimplement a bunch of stuff i'm too tired
           to :/.

           These tags have to be treated like block markup so that they
           work for the common case that they wrap {:KEYWORD:}
           tags. In order to simplify that process, we transform the
           '***', '**', and '*' markup into {{bi}}, {{b}}, {{i}},
           respectively, so that it's in a format p10BreakingBlocks()
           can deal with.
        */
        const p09MiniMarkdownKludge = function(){
            affirm(this.top && 1===this.top.length,
                   "Expecting length-1 array this.top.");
            var body = this.top.pop();
            const self = this;
            const doIt = function f(rx, tag){
                const split = body.split(rx);
                if(0===split.length%2){
                    self.error("Mismatched '"+tag+"' count. Skipping",body);
                    return f;
                }
                const buf = [],
                      n = split.length,
                      mark = (x)=>"{{"+(x%2 ? "/":"")+tag+"}}";
                var i = 0, alt = 0;
                for( ; i < n; ++i ){
                    const p = split[i];
                    if(i) buf.push(mark(alt++));
                    pushNoFalsy(buf, p);
                }
                body = buf.join('');
                return f;
            };
            body = body.split(/\\\*/).join('{{asterisk}}');
            doIt(/\*{3}/g, 'bi')(/\*{2}/g, 'b')(/\*/g,'i');
            // we can't do (/_/g,'u') here because that breaks
            // KEYWORDS_WITH_UNDERSCORES.
            this.top.push(body);
        };

        /**

         */
        const p05ObligatoryNewlineKludges = function(){
            affirm(this.top && 1===this.top.length,
                   "Expecting length-1 array this.top.");
            this.top[0] = this.top[0].replace(
                /(\{\{\/[^}]+}})\r?\n/g, '$1'
                /**
                   Remove a \n immediately after block markup, else
                   we get huge gaps between, e.g.:

                   {{colorbox...}}...{{/colorbox}}
                   {{colorbox...}}...{{/colorbox}}

                   Block constructs force a line break to their left,
                   anyway, so the \n is superfluous (and problematic)
                   for output purposes.
                */
            );
        };
        
        /**
           Dives into a MarkerUpper's content, recursing into
           the body text of each block-level element it discovers,
           replacing it with the DOM element returned by that
           recursion.
        */
        const p10BreakingBlocks = function f(){
            affirm(this.top && 1===this.top.length,
                   "Expecting length-1 array this.top.");
            const opener = '{{';
            if(!this.isRecursing){
                /* ^^^ if we're recursing, this part has already been
                   done from the top-most call. */
                p05ObligatoryNewlineKludges.call(this)/*unfortunate pre-processing*/;
                p09MiniMarkdownKludge.call(this)/*unfortunate pre-processing*/;
            }
            affirm(this.top && 1===this.top.length,
                   "Expecting (again) length-1 array this.top.");
            if(!f.rx){
                f.rx = /\{\{(\!|\.|[-\w]+)/g;
            }
            f.rx.lastIndex = 0;
            const self = this;
            const body = this.top.pop();
            var prevLastIndex = 0,
                m;
            while(m = f.rx.exec(body)){
                const tag = m[1],
                      tStart = f.rx.lastIndex - m[0].length,
                      isComment = '!'===tag[0];
                const tagHandler = (
                    isComment ? undefined : this.config.TagHandlers.getHandler(tag)
                );
                if(!tagHandler && !isComment){
                    this.error("No handler found for tag type '"+tag+"'. SKIPPING.");
                    continue;
                }
                if(isComment){
                    // It seems we're not going to get around needing this feature :/.
                    const tagClose = '!}}',
                          tOpenEnd = f.rx.lastIndex,
                          commentEnd = body.indexOf(tagClose, tOpenEnd);
                    if(commentEnd<0){
                        this.toss("Cannot find closing '!}}' for comment "
                                  +"tag starting at character offset "
                                  +f.rx.lastIndex+".");
                    }
                    pushNoFalsy( // Flush leading non-markup part...
                        this.top,
                        body.substring(prevLastIndex, tStart)
                    );
                    prevLastIndex = f.rx.lastIndex = commentEnd + 3;
                    //console.debug({comment: body.substring(tOpenEnd, commentEnd)});
                    continue;
                }/* end {{! ... !}} handling */

                var tOpenEnd = body.indexOf('}}', f.rx.lastIndex) + 2/*skip '}}'*/;
                if(tOpenEnd<2){
                    this.toss("Cannot find '}}' closer for '"+tag+"' tag.");
                }
                else if(this.inlineOnly && tagHandler.isBlock){
                    this.toss("Block-level markup is not permitted here.");
                }else if(0){
                    /* Convenience workaround for JSON objects as tag
                       parameters, to keep them from being forced to
                       add a space after the objects. This still does
                       not allow nested objects without intervening
                       spaces, though. e.g.:

                       {{tag {foo:{bar:a}},baz:3}}}

                       will see the '}}' at 'bar:a}}' as the end of
                       the tag, so that would would require a space
                       between the braces.
                    */
                    while('}'===body[tOpenEnd]) ++tOpenEnd;
                }
                const tagArgs = body.substring(f.rx.lastIndex, tOpenEnd-2).trim();
                const tagClose = tagHandler.hasClosingTag ? "{{/"+tag+"}}" : undefined,
                      innerEnd = tagClose
                      ? body.indexOf(tagClose, tOpenEnd)
                      : tOpenEnd;
                //console.debug("tagClose=",tagClose,"innerEnd=",innerEnd);
                if(innerEnd<0){
                    this.error("Skipping: unmatched {{"+tag+"}} at character offset "
                               +tStart+".");
                    console.error(
                        "More info about skipped",tag,"tag:",{
                            tagHandler,
                            "f.rx.lastIndex":f.rx.lastIndex,
                            tOpenEnd, tagArgs, tagClose,
                            "body starting at f.rx.lastIndex": body.substr(f.rx.lastIndex)
                        }
                    );
                    continue;
                }
                if(0) console.debug("innerEnd =",innerEnd,
                                    "tStart =", tStart,
                                    "m[0].length =",
                                    m[0].length);
                pushNoFalsy(// Flush initial non-block part...
                    this.top,
                    body.substring(prevLastIndex, tStart)
                );
                var tmp;
                const inner = tagClose
                      ? (
                          (tmp=body.substring(
                              tOpenEnd, innerEnd
                          ))
                              ? (
                                  tagHandler.trimBodyMethod ? tmp[tagHandler.trimBodyMethod]() : tmp
                              ) : tmp
                      ) : undefined;
                tmp = undefined;
                const markupResult = tagHandler.process(
                    this, tag,
                    preprocessTagArgs(this, tag, tagHandler, tagArgs),
                    processTagBody(this, tag, tagHandler, inner)
                );
                if(markupResult){
                    if(markupResult.forEach){
                        markupResult.forEach((v)=>pushNoFalsy(this.top, v));
                    }else{
                        this.top.push(markupResult);
                    }
                }
                if(0) console.debug("Tag =",tag,", tagArgs =",tagArgs,
                                    "start/end/lastIndex =",
                                    tStart, innerEnd, f.rx.lastIndex,
                                    "inner =(((",inner,")))");
                prevLastIndex = innerEnd + (tagClose ? tagClose.length : 0);
                if(0) console.debug(tag, "Next few chars <<<",body.substr(prevLastIndex, 15),'>>>');
                let foundNL = false, charCode;
                f.rx.lastIndex = prevLastIndex;

                if(0) console.debug("At prevLastIndex",prevLastIndex,"=",
                                    body.substr(prevLastIndex,25),"...");
            }
            // Any remaining non-block part...
            pushNoFalsy( this.top,
                         body.substr(prevLastIndex) );
            //console.debug("p10BreakingBlocks(): this.top =",this.top);
            return this;
        };

        const p99Dom = function(){
            const d = dom.create(
                this.domTagType || 'div' /*(
                                           this.inlineOnly ? 'span' : 'div'
                                           // ^^^ currently breaks card header layout
                                           )*/
            );
            affirmArray(this.top, "Expecting array this.top.");
            this.top.forEach((p)=>{
                if(isString(p)) p = dom.text(p);
                d.appendChild(p);
            });
            this.domElement = d;
            return this;
        };

        MarkerUpper.dom = DV.dom;
        
        MarkerUpper.prototype = {
            /**
               Regexes which are commonly used by markup tag handlers.
            */
            /**
               Cached common regexes
            */
            rx: {
                /**
                   Matches dv-conventional font size CSS class names,
                   with the part after "dv-font-size-" explicitly being a
                   capture group.
                */
                fontSizeClass: /\bdv-font-size-([-\w]+)/,
                // "Global" match equivalent of fontSizeClass.
                fontSizeClass_g: /\bdv-font-size-([-\w]+)/g,
                /** Regex for splitting a string on multiple spaces. */
                splitSpaceDelim: /\s+/g,
                colorClass: /\bdv-fg-[-\w]+/g,
                bgClass: /\bdv-bg-[-\w]+/g
            },
            config:{
                /**
                   All but the lowest-level markup tag handlers. A few
                   are handled in a pre-processing phase which is not
                   represented in this object.
                */
                TagHandlers,

                /**
                   Default language designation for keyword
                   translations.
                */
                language: 'en',
                /**
                   Various CSS classes used by markup output.
                */
                css:{
                    boldItalic: 'italic strong',
                    bold: 'strong',
                    italic: 'italic',
                    cardGlyph: 'dv-glyph',
                    /** Lowest dv-font-size-XX value. */
                    minFontSize: 50,
                    /** Highest dv-font-size-XX value. */
                    maxFontSize: 250,
                    /** The dv-font-size-XX value at which the CSS
                        switches from 1% font size granularity to 5%.
                    */
                    granularityThreshold: 250,
                    keywordRef: 'dv-kwref'
                },
                /**
                   The top-most URI path to this library's assets,
                   INCLUDING a trailing slash (as that simplifies its
                   usage). If this file is included from the same
                   directory as the HTML file which includes it, this
                   will be set to the empty string. This string is
                   used to modify the paths to the configured glyphs
                   which this class emits, prepending this path to
                   their configured path.
                */
                libDirUri: (function(){
                    // Determine URI of this script as it's being initially run...
                    // https://elijahmanor.com/can-a-javascript-file-tell-what-url-requested-itself/
                    // Trivia: it's the same approach that Makefiles use..
                    var currentScript;
                    if(document.currentScript){
                        currentScript = document.currentScript.src;
                    }else{
                        let scriptTags = document.getElementsByTagName( "script" );
                        currentScript = scriptTags[scriptTags.length - 1].src;
                    }
                    const a = currentScript.split('/');
                    if(1===a.length) return '';
                    a.pop();
                    const uri = a.join('/')+'/';
                    return uri;
                })(),
                /**
                   Keywords and standard glyphs. The img properties must point
                   to a directory with "/black/" in the name so that
                   we have a standard way to generate other color variants
                   (by replacing /black/ with /colorname/).

                   Hungarian translations provided by
                   Andris Borbíró (https://www.boardgamegeek.com/user/ramiz).

                   Labels may be one of:

                   - A string - assumed to be the English label.

                   - A map of language codes to strings. The language
                   code defined in thisMarkupParser.config.language is
                   used, if available, falling back to 'en'.

                   - Falsy, to indicate that the entry represents only
                   a glyph with no standard label.

                   The 'ddicon' property is intended to replace the
                   'img' property. Its value is a key for use with
                   ddIconFontEntries, which maps the icon to a font
                   entry.

                   Potential TODO: remove the 'glyphs/' parts from all of the
                   names, as they're the same for all images. OTOH, they might
                   not *stay* the same for all images.
                */
                keywords:{
                    1:{ label:'', img: "glyphs/black/die-1.svg" },
                    2:{ label:'', img: "glyphs/black/die-2.svg" },
                    3:{ label:'', img: "glyphs/black/die-3.svg" },
                    4:{ label:'', img: "glyphs/black/die-4.svg" },
                    5:{ label:'', img: "glyphs/black/die-5.svg" },
                    6:{ label:'', img: "glyphs/black/die-6.svg" },
                    "F_ARISTOCRACY": {
                        label: "Aristocracy",
                        img: "glyphs/factions/aristocracy.svg"
                    },
                    "F_BEAST": {
                        label: "Beast",
                        img: "glyphs/factions/beast.svg"
                    },
                    "F_BLOODSMAN": {
                        label: "Bloodsman",
                        img: "glyphs/factions/bloodsman.svg"
                    },
                    "F_CELESTITE": {
                        label: "Celestite",
                        img: "glyphs/factions/celestite.svg"
                    },
                    "F_CONSTRUCT": {
                        label: "Construct",
                        img: "glyphs/factions/construct.svg"
                    },
                    "F_DIAMOND": {
                        label: "Diamond",
                        img: "glyphs/factions/diamond.svg"
                    },
                    "F_GHOST": {
                        label: "Ghost",
                        img: "glyphs/factions/ghost.svg"
                    },
                    "F_HUNTSMAN": {
                        label: "Huntsman",
                        img: "glyphs/factions/huntsman.svg"
                    },
                    "F_ORPAL": {
                        label: "Orpal",
                        img: "glyphs/factions/orpal.svg"
                    },
                    "F_OVOID": {
                        label: "Ovoid",
                        img: "glyphs/factions/ovoid.svg"
                    },
                    "F_PLANT": {
                        label: "Plant",
                        img: "glyphs/factions/plant.svg"
                    },
                    "F_VARPEN": {
                        label: "Varpen",
                        img: "glyphs/factions/varpen.svg"
                    },
                    "F_WARLOCK": {
                        label: "Warlock",
                        img: "glyphs/factions/warlock.svg"
                    },
                    HEALTH:{
                        label:"",
                        img: "glyphs/black/health.svg"
                    },
                    HEALTH_BASE:{
                        label:"",
                        img: "glyphs/black/health-base.svg"
                    },
                    LUCK:{
                        label:"Luck",
                        img: "glyphs/black/attr-luck.svg"
                    },
                    MIND:{
                        label:"Mind",
                        img: "glyphs/black/attr-mind.svg"
                    },
                    POWER:{
                        label:"Power",
                        img: "glyphs/black/attr-power.svg"
                    },
                    SPEED:{
                        label:"Speed",
                        img: "glyphs/black/attr-speed.svg"
                    }
                }
            },
            
            /**
               May be replaced by client apps.
            */
            error: function(){
                console.error.apply(console, argsToArray(arguments));
            },
            /**
               Reports its arguments using this.error() and then
               throws. If overridden, any override MUST throw - the
               internals of the markup handler rely on that.
            */
            toss: function(){
                const args = argsToArray(arguments);
                args.unshift("MarkerUpper:");
                this.error.apply(this, args);
                throw new Error( args.join(' ') );
            },

            /**
               Resets this object's parsing state.
            */
            reset: function(){
                delete this.content;
                delete this.top;
                delete this.domElement;
                return this;
            },

            /**
               Returns the URI to the "glyphs" directory, including the
               trailing slash. If passed an argument, it returns that path
               prepended to the argument.
            */
            getGlyphDirUri: function(colorAndName){
                return this.config.libDirUri+'glyphs/'+(arguments.length ? arguments[0] : '');
            },

            /**
               Equivalent to calling this.getGlyphDirUri() and passing
               it colorAndName+'.svg', except that if colorAndName has
               no '/' then it is assumed to be only a name, without a
               color part, and the color is derived from
               this.styleContext.fg, defaulting to 'black'.
            */
            getGlyphUri: function(colorAndName){
                var arg;
                if(colorAndName.indexOf('/')>=0){
                    arg = colorAndName;
                }else{
                    arg = ((
                        this.styleContext ? this.styleContext.fg : 0
                    ) || 'black')+'/'+colorAndName;
                }
                return this.getGlyphDirUri(arg+'.svg');
            },

            /**
               Expects colorAndName to be a string in the form
               "color/glyphName" or "glyphName" (with no slash).
               For the former case it returns the "color" part of
               the string. For the latter it returns the undefined
               value (which *normally*, but not always, means "black"
               was intended).
            */
            getColorFromGlyphUri: function(colorAndName){
                const x = colorAndName.split('/',2);
                return x.length>1 ? x.shift() : undefined;
            },

            /**
               Given a color, which is intended to be a
               "regional" color, this function returns
               the name for the complementary color intended
               for use with the "X" cross-out used by Lairs and
               Encounter card sector headers. This is currently red for all colors
               except for red and magenta, which return 'yellow'.
            */           
            getXColorForUnderColor: function f(underColor){
                if(!f.map){
                    f.map = {
                        red: 'yellow',
                        law: 'yellow',
                        magenta: 'yellow'
                    };
                }
                return f.map[underColor] || 'red';
            },
            
            /**
               If txt is truthy, it is assumed to be a string, it is
               split() on \s+, and the result is returned. If it
               falsy, an empty array is returned.
            */
            splitSpaceDelim: function(txt){
                return txt ? txt.split(this.rx.splitSpaceDelim) : [];
            },

            /**
               The framework-conventional color names, noting that these
               names refer to lower-level mappings of the color codes,
               not necessarily a commonly-understood color value.

               Caveats:

               - The "real" list of colors is defined via CSS, so this
               method is only "correct" insofar as it stays in sync
               with the CSS.

               - The names "default" and "inherit" do not have color
               codes associated with them. They are used in certain
               contexts to tell HTML elements to use a corresponding
               color (foreground or background, depending on the
               context) derived from various CSS-side values or to
               inherit the corresponding color the containing HTML
               element.

               - The name "override" is a very special case for use
               with certain CSS selectors, to allow overridden colors
               to be applied. As of this writing it's not actually
               used, but a proof-of-concept test was made. It's not
               yet clear whether we could actually make good use of
               this.

               Modifying this list at runtime will very likely not
               have any useful effect.
            */
            frameworkColorNames: [
                "black", "blue", "default", "green",
                "green-se", "grey", "inherit", "magenta",
                "orange", "purple", "red", "white", "yellow"
                // "override" /* very special case */
            ],
            /**
               Returns true if this class believes c to be a
               framework-conventional color name, else false.

               Caveats:

               The "real" list of colors is essentially defined via
               CSS, so this method is only "correct" insofar as it
               stays in sync with the CSS.
            */
            isColorName: function f(c){
                if(!f.map){
                    f.map = {};
                    this.frameworkColorNames.forEach((c)=>f.map[c]=undefined);
                }
                return f.map.hasOwnProperty(c);
            },

            /**
               Given a known color scheme name, this function returns
               the "preferred" color of glyphs for rendering on that
               scheme's default background color. Returns dflt if no
               match is found.

               FIXME: we really need 2 levels of mapping: one for the
               default bg color and one for an arbitrary user-supplied
               color (that mapping could be generic, overridden by
               colorscheme-specific values). In order to be able to
               use that from the markup layer, we need a way to
               communicate the current card object to the markup
               parser, which requires a bit of API surgery.
            */
            deriveGlyphColor: function f(colorScheme, dflt){
                if(!f.gColorMap){
                    f.gColorMap = {
                        /* Default glyph colors, depending on colorscheme.
                           This need to jive with the scheme-specific background
                           color CSS or Grief will likely ensue. */
                        wetlands:"blue",
                        highlands:"yellow",
                        badlands:"red",
                        lowlands:"blue",
                        settlement:"green-se",
                        tundra:"purple",
                        seventies:"black",
                        law:"magenta"
                    }
                }
                return (
                    f.gColorMap.hasOwnProperty(colorScheme)
                        ? f.gColorMap[colorScheme || dflt]
                        : dflt
                );
            },
            
            /**
               Returns n if the integer number n represents a "legal"
               dv-font-size-XX value, else returns 0. The legal range of
               values is taken from this framework's CSS code, and
               needs to be kept in sync with that.
            */
            isLegalFontSizeHint: function(n){
                if(isNaN(n)) return 0;
                const C = MarkerUpper.prototype.config.css;
                n = Math.floor(n/*noting that floor() converts numeric strings*/);
                if(n>=C.minFontSize && n<=C.granularityThreshold) return n;
                else if(n<=C.maxFontSize && !(n%5)) return n;
                return 0;
            },

            /**
               Given a framework-conventional font size integer value
               and a number of sizes to "step" that value, returns the
               next "legal"[1] font size. The stepCount may be
               negative, meaning to step that many ticks
               downwards. The exact value which corresponds to a
               "step" depends on whether the currentSize is within the
               framework's "high-res" range (1% granularity) or "low-res"
               range (5% granularity). e.g. passing (250, 1) will return
               255[2], but passing (199,1) will return 200. Likewise,
               (300,-5) would return 275, but (199,-5) would result in 194.

               A stepCount of 0 is a special case: the first argument
               will be checked to confirm that it is in range. If it
               is in the 5%-granularity range but is not an even multiple
               of that granularity, it will be rounded down to the next
               even multiple. e.g. passing in (312, 0) will return 310
               and passingin (X,0), where X is lower than the minimum or
               higher than the maximum will return the minimum resp.
               maximum value. Passing in (112,0) will return 112 because
               it's a legal value as-is.

               If the new font size would be lower or higher than the
               configured minimum/maximum, the minimum resp. maximum
               configured value is returned.

               [1]: "legal" means that we have a corresponding
               dv-font-size-XX value.

               [2]: assuming that (this.config.css.granularityThreshold===250).
               If that value is changed then the above example may be
               invalid.
            */
            stepFontSize: function f(currentSize, stepCount){
                if(!f.C){
                    f.C = MarkerUpper.prototype.config.css;
                }
                currentSize = +currentSize;
                stepCount = +stepCount;
                var stepSize = 0;
                if(stepCount){
                    if(currentSize === f.C.granularityThreshold){
                        stepSize = stepCount<0 ? 1 : 5;
                    }else if(currentSize < f.C.granularityThreshold){
                        stepSize = 1;
                    }else{
                        stepSize = 5;
                    }
                }
                const newSz = currentSize + (stepCount * stepSize);
                if(newSz <= f.C.minFontSize){
                    return f.C.minFontSize;
                }else if(newSz >= f.C.maxFontSize){
                    return f.C.maxFontSize;
                }
                else if(newSz <= f.C.granularityThreshold){
                    return newSz;
                }else{
                    return newSz - (newSz % 5);
                }
            },
            
            /**
               Removes any class(es) matching this.rx.fontSizeClass_g
               from the given DOM element then, if n is truthy...

               Replaces any existing dv-font-size-XXX class in DOM
               element e with the given font size, which may either be
               an integer with a framework-conventional value (this
               function does not confirm its validity) or a string,
               which is assumed to be a dv-font-size-XXX CSS class
               which applies the size (it is applied as-is).

               Returns this object.
            */
            setFontSizeHint: function(e, n){
                e.className = e.className.replace(this.rx.fontSizeClass_g, '');
                if(n){
                    if('number'===typeof n) e.classList.add('dv-font-size-'+n);
                    else /*assume fully-qualified value*/e.classList.add(n);
                }
                return this;
            },

            /**
               Calls Removes any CSS class(es) matching
               this.rx.colorClass, then, if n is truthy...

               If n matches this.rx.colorClass then n is added to e's
               classList as-is, else ('dv-fg-'+n) is added.

               Returns this.
            */
            setColorClass: function(e, n){
                e.className = e.className.replace(this.rx.colorClass, '');
                if(n){
                    if(this.rx.colorClass.test(n)) e.classList.add(n);
                    else e.classList.add('dv-fg-'+n);
                }
                return this;
            },

            /**
               The dv-bg-COLOR counterpart of setColorClass().
            */
            setBgClass: function(e, n){
                e.className = e.className.replace(this.rx.bgClass, '');
                if(n){
                    if(this.rx.bgClass.test(n)) e.classList.add(n);
                    else e.classList.add('dv-bg-'+n);
                }
                return this;
            },

            /**
               Runs the markup parser on the given input text and
               returns a new DOM element containing the text.  If the
               input is falsy (empty) then the undefined value is
               returned. Throws on error. It will typically return a
               DIV element, but certain uses may cause it to return
               another element type.

               If this.inlineOnly is truthy then only "inline"
               parsing is performed: any block elements will be
               treated as an error. This is intended primarily for use
               in rendering keyword glyphs in the "type" field of
               certain cards.
            */
            markup: function(input){
                affirmString(input,"Non-string provided as markup() input.");
                this.content = input.trim();
                this.top = undefined;
                this.domElement = undefined;
                if(!this.content) return this.domElement;
                try {
                    // Bootstrapping...
                    p00Escapes.call(this);
                    /**
                       Because of how we recurse and hold our output
                       state, we can't simply process block-level
                       markup in any order.  We have to process them
                       in the order they are found in the content,
                       so...

                       Note that we run this part in inlineOnly mode
                       only to handle the italics/bold markup. If any
                       "incompatible" markup is entired in this mode,
                       an exception will be thrown.
                    */
                    p10BreakingBlocks.call(this);

                    // Non-recursive markup...
                    if(this.inlineOnly){
                        p25Keywords.call(this);
                    }else{
                        p20Separators.call(this);
                        p25Keywords.call(this);
                        p50AttrMod.call(this)
                        /* attribute modifiers "technically" inline,
                           but we don't treat them as such here
                           because they have no place being used in
                           the contexsts we expect to use inlineOnly
                           mode for (the "type" field of certain
                           mini-sized card types, and potentially for
                           enemy attributes in the footers of
                           Encounter-style cards).*/;
                    }
                    if(0 && !this.isRecursing){
                        console.debug("Final this.top =",this.top);
                    }
                    p99Dom.call(this);
                }catch(e){
                    this.reset();
                    throw e;
                }
                const rc = this.domElement;
                this.reset();
                return rc;
            }

        };

        return MarkerUpper;
    })(global.DarkVenture) /* end MarkerUpper module */;

    /* Set up most of the DarkVenture object/APIs... */
    (function(DV/*==>global.DarkVenture*/){
        const dom = DV.dom;

        /**
           If x is a string or falsy, it is returned as-is, else...

           If it is an array, the result of x.join('') is
           returned, else...

           If it is an object, this function returns the result of
           passing thatObject.text to this function, else...

           console.warn() is called and x is returned as-is.
        */
        const getCardBodyText = DV.getCardBodyText = function f(x){
            if(!x || 'string' === typeof x) return x;
            else if(x instanceof Array) return x.join('');
            else if(x.hasOwnProperty('text')) return f(x.text);
            else {
                console.warn("Unhandled value passed to getCardBodyText()", x);
                return x;
            }
        };

        const getCardBodyFontSize = function f(x){
            if('number' === typeof x) return 'dv-font-size-'+x;
            else if(!x || 'string' === typeof x || x instanceof Array) return undefined;
            else if(x.hasOwnProperty('fontSize')) return f(x.fontSize);
            else {
                //console.warn("Unhandled value passed to getCardBodyFontSize()", x);
                return undefined;
            }
        };

        /**
           Assumes that x is a card body object or card body text (string or array),
           and returned a normalized object in the form {text: X, fontSize: Y}, where
           X is the body text (as a string) and Y is either a font size or the undefined
           value. */
        const simCardBody = function(x){
            return {
                text: getCardBodyText(x),
                fontSize: getCardBodyFontSize(x)
            };
        };
        
        DV.config = {
            libDirUri: (function(){
                // Determine URI of this script as it's being initially run...
                // https://elijahmanor.com/can-a-javascript-file-tell-what-url-requested-itself/
                // Trivia: it's the same approach that Makefiles use..
                var currentScript;
                if(document.currentScript){
                    currentScript = document.currentScript.src;
                }else{
                    let scriptTags = document.getElementsByTagName( "script" );
                    currentScript = scriptTags[scriptTags.length - 1].src;
                }
                const a = currentScript.split('/');
                if(1===a.length) return '';
                a.pop();
                const uri = a.join('/')+'/';
                return uri;
            }),
            cssClass: {
                cardWrapperMini: 'dv-card-wrapper-mini',
                cardWrapperStandard: 'dv-card-wrapper-standard',
                dvCard: 'dv-card',
                cardSizeMini: 'dv-card-size-mini',
                cardSizeStandard: 'dv-card-size-standard',
                cardSQ: 'dv-card-sidequest',
                cardHQ: 'dv-card-heroicquest'
            }
        };
        /**
           Cached common regexes
        */
        DV.rx = {
            /**
               Matches morbad-conventional font size CSS class names,
               with the part after "dv-font-size-" explicitly being a
               capture group.
            */
            fontSizeClass: /\bdv-font-size-([-\w]+)/,
            // "Global" match equivalent of fontSizeClass.
            fontSizeClass_g: /\bdv-font-size-([-\w]+)/g,
            /** Regex for splitting a string on multiple spaces. */
            splitSpaceDelim: /\s+/g,
            colorClass: /\bdv-fg-[-\w]+/g,
            bgClass: /\bdv-bg-[-\w]+/g
        };

        /**
           Expects cssClass to either be a string (space-delimited CSS
           class name list), an array of CSS class names (1 class per
           array entry), or a falsy value.

           It returns a normalized form of the value: a string by default
           (suitable for use as a CSS class list property) or an array if
           asArray is truthy. For falsy values it returns an empty string
           resp. empty array.
        */
        DV.cssClassVal = function(cssClass,asArray){
            if('string' === typeof cssClass){
                return asArray ? dom.splitClassList(cssClass) : cssClass;
            }else if(cssClass instanceof Array){
                return asArray ? cssClass : cssClass.join(' ');
            }else{
                return asArray ? [] : '';
            }
        };

        /**
           Expects e to be a DOM element and filters to be an array
           in the form:

           ["css-filter-name", filter value, ..."css-filter-name-N",
           filter value N]

           It requires that each filter name is one of:

           blur, brightness, contrast, grayscale, hue-rotate,
           invert, opacity, saturate, sepia           

           And that the value following the filter name is one
           appropriate for use with the given CSS filter name. e.g.
           ["grayscale", "75%"] or ["grayscale", 75] are equivalent
           (see below for why). This routine does not know which
           values are legal for any given filter, so does validate
           them.

           That said... if passed a numeric value >1.0 for a given
           filter, this function converts it to a string and appends
           a suffix to it depending on the filter name:

           greyscale, invert, sepia, saturate, contrast, brightness,
           opacity: "%"

           hue-rotate: "deg"

           If passed a filter not in those lists, the value passed in
           is not treated specially. e.g. blur requires a "distance"
           and this function does not assume any given measurement unit.

           If passed a number <= 1.0 it is assumed to be a
           fraction, and many filters treat 50% and 0.5 equally.

           This admittedly unusual structure is necessary because the
           order of application of filters is significant and an array
           is the only structure which guarantees us, in a
           JSON-friendly way, a well-defined order.

           The filter/value combinations are combined into a single
           "filter" style property value and assigned to
           e.style.filter.

           CSS filter docs:

           https://developer.mozilla.org/en-US/docs/Web/CSS/filter
        */
        DV.applyCssFilters = function f(e, filters){
            if(!f.hasOwnProperty('filterValue')){
                f.map = {
                    "hue-rotate": 'deg'
                };
                [
                    'grayscale',
                    'invert',
                    'sepia',
                    'saturate',
                    'contrast',
                    'brightness',
                    'opacity'
                ].forEach((v)=>f.map[v]='%');
                f.filterValue = function(k,v){
                    return (
                        ('number'===typeof v && v>1.0)
                            ? ''+(v|0)+f.map[k]
                            : v
                    );
                };
            }
            const s = []
            filters.forEach(function(v,n){
                if(n%2) return /* odd-numbered elements */;
                s.push([v,'(', f.filterValue(v,filters[n+1]),
                        ')'].join(''))
            });
            e.style.filter = s.join(' ');
        };

        ['setFontSizeHint','setColorClass','setBgClass'
        ].forEach((m)=>DV[m] = DV.MarkerUpper.prototype[m]);

        /**
           Usages:

           1) (tgtElem, text [, inlineOnly=false[,language='en']])

           Pass in a target DOM element and markup text, and it will
           mark up the text and replace the DOM element's contents with
           it. Returns the target element.

           2) (text)

           Pass in a markup text, and it will mark up the text and
           return the output (a DOM element or, for empty input, the
           undefined value).

           3) (falsy, text [, inlineOnly=false [,language='en']])

           Processes the given markup text. If tgtElem is truthy,
           the resulting marked-up content replaces tgtElem's contents,
           otherwise a new DIV DOM element is created to hold the contents
           and it is returned.

           If text is falsy or empty (containing only spaces) then: if
           tgtElem is not falsy, it is emptied and returned, else an
           empty Text DOM node is returned.

           If inlineOnly is truthy then block-style markup is not
           permitted (and will trigger an exception).
        */
        DV.runDVMarkup = function f(tgtElem, text, inlineOnly, language){
            if(undefined===f.isRunning) f.isRunning = 0;
            if(1===arguments.length){
                text = arguments[0];
                tgtElem = undefined;
            }
            if(!text || !(text = text.trim())){
                if(tgtElem) return dom.clearElement(tgtElem);
                else return dom.text('');
            }
            const m = new DV.MarkerUpper(!!f.isRunning++);
            m.config.language = language || 'en';
            m.inlineOnly = !!inlineOnly;
            var marked;
            try{
                marked = m.markup(text);
            }finally{
                --f.isRunning;
            }
            return tgtElem
                ? dom.moveChildrenTo(dom.clearElement(tgtElem),marked)
                : marked;
        },
        DV.runDVMarkup.isRunning = 0;

        /**
           For each DOM element matching the given selector, this
           routine extracts that element's innerText, uses it as input
           for runDVMarkup(), and replaces the contents of that
           element with the result of that function.

           The selector may be a CSS selector string, a single
           HTMLElement, or a container type with a forEach() method.

           If classToRemove is truthy then it is assumed to be a CSS
           class which should be removed from each element as it is
           processed (to avoid duplicate processing later, which
           "would be bad").

           Exceptions while running markup are propagated.

           Returns this object.

           See also: applyMarkupToTextNodes()

           PS: in hindsight, this function probably isn't really
           needed, as applyMarkupToTextNodes() does essentially the
           same thing but in a way which (currently, at least) better
           fits our needs.
        */
        DV.applyMarkupToElements = function(selector,classToRemove){
            const selection = ('string' === typeof selector) ?
                  document.querySelectorAll(selector)
                  : (selector instanceof HTMLElement
                     ? [selector] : selector);
            const self = this;
            selection.forEach(function(e){
                if(classToRemove) dom.removeClass(e, classToRemove);
                self.runDVMarkup(e, e.innerText);
            });
            return this;
        };

        /**
           For each DOM element matching the given selector, this
           routine iterates over all of its children
           (non-recursively). Those which are not of type TEXT_NODE
           are not modified. Those which are of type TEXT_NODE and
           contain non-whitespace content are replaced by the results
           of passing that text to runDVMarkup().

           The selector may be a CSS selector string, a single
           HTMLElement, or a container type with a forEach() method.

           If classToRemove is truthy then it is assumed to be a CSS
           class which should be removed from each element as it is
           processed (to avoid duplicate processing later, which
           "could be bad").

           Exceptions while running markup are propagated.

           Returns this object.

           See also: applyMarkupToElements()
        */
        DV.applyMarkupToTextNodes = function f(selector,classToRemove){
            if(!f.doChild){
                /* Reminder to self: this "static" function is only
                   possible because the morbad object is a singleton,
                   so "self" will always been the same instance. */
                const self = this;
                f.doChild = function(child, parent /*argument count/order is important!*/){
                    const txt = child.nodeValue.trim();
                    if(txt){
                        const m = self.runDVMarkup(undefined,txt);
                        while(m.firstChild) parent.insertBefore(m.firstChild, child);
                        parent.removeChild(child);
                    }
                };
            }
            const selection = ('string' === typeof selector) ?
                  document.querySelectorAll(selector)
                  : (selector instanceof HTMLElement
                     ? [selector] : selector);
            selection.forEach(function(parent){
                dom.visitTextNodes(parent, f.doChild);
                if(classToRemove) dom.removeClass(parent, classToRemove);
            });
            return this;
        };

        /**
           Given a single DOM element, this function *recursively*
           passes each ELEMENT_NODE-type node to
           applyMarkupToTextNodes().

           If the 2nd argument is truthy, the text nodes are processed
           unconditionally, else they are processed only for DOM nodes
           with the CSS class 'markup'. In either case, that CSS class
           is removed from each node as it is processed, to avoid
           potential duplicate processing later on.

           Any nodes which are not of type ELEMENT_NODE are ignored
           by this routine.

           Note that, unlike applyMarkupToTextNodes(), this routine
           accepts only a single DOM node, not a selector or list of
           nodes, for sanity's sake (to avoid potentially traversing
           the same nodes multiple times via the recursion).

           Returns this object.
        */
        DV.applyMarkupToTree = function f(root, unconditional){
            if(root.ELEMENT_NODE !== root.nodeType) return this;
            if(!f.cssClass) f.cssClass = 'markup';
            const self = this;
            if( root.childNodes.length ){
                root.childNodes.forEach(function(ch){
                    if(ch.ELEMENT_NODE === ch.nodeType){
                        f.call(self, ch, unconditional);
                    }
                    /* TEXT_NODEs are handled in the next step, and we want to skip
                       COMMENT_NODEs and such. */;
                });
            }
            if( unconditional || root.classList.contains( f.cssClass ) ){
                this.applyMarkupToTextNodes(root, f.cssClass);
            }
            return this;
        };

        /**
           Retuns a UUIDv4-style string, but filled entirely with
           random digits, not with two fixed digits like
           UUIDv4. Intended for use in generating IDs for fan-created
           cards.
        */
        DV.randomId = function(){
            return 'xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'.replace(/x/g, function(c) {
                return (Math.random() * 16 | 0).toString(16);
            });
        };

        /**
           Various internal card rendering utilities used by multiple card
           classes in their JSON==>HTML conversions.
        */
        const CardRenderUtil = {
            markup: function(eDest, txt, inlineOnly){
                return DV.runDVMarkup(eDest, txt, inlineOnly);
            },
            inlineMarkup: function(eDest, txt){
                return this.markup(eDest, txt, true);
            },
            setFg: function(e, j, dflt/*"default" || "inherit"*/){
                if(j.fg || dflt) DV.setColorClass(e, j.fg || dflt);
                return this;
            },
            setBg: function(e, j, dflt/*"default" || "inherit"*/){
                if(j.bg || dflt) DV.setBgClass(e, j.bg || dflt);
                return this;
            },
            setColors: function(e, j, dflt/*"default" || "inherit"*/){
                return this.setFg(e,j,dflt).setBg(e,j,dflt);
            },
            setFontSize: function(e, j){
                if(j.fontSize) DV.setFontSizeHint(e, j.fontSize);
                return this;
            },
            setCSS: function(e, j){
                if(j.cssClass) DV.dom.addClass(e, j.cssClass);
                if(j.cssFilter) DV.applyCssFilters(e, j.cssFilter);
                return this;
            },
            setStyle: function(e, j,defaultColors){
                this.setCSS(e,j).
                    /* do setCSS() first so that if it contains
                       dv-fg-XXX/dv-bg-XXX/dv-font-size-XXX, setColors()
                       and setFontSize() can behave properly (overwrite
                       it). If we do cssClass last, we can possibly end up
                       with multiple such conflicting styles. */
                setFontSize(e,j).setColors(e,j,defaultColors);
                return this;
            }
        }/*CardRenderUtil*/;

        
        const SQCard = DV.SQCard = function(json){
            if(json){
                this.populateFromJSON(json);
            }else{
                this.resetDOMElement();
            }
        };
        SQCard.css = {
            cardSize: DV.config.cssClass.cardSizeMini,
            cardType: DV.config.cssClass.cardSQ,
            cardWrapper: DV.config.cssClass.cardWrapperMini,
            scriptTagSelector: 'script[type="x-dv-card-sidequest"]'
        };
        const HQCard = DV.HQCard = function(json){
            if(json){
                this.populateFromJSON(json);
            }else{
                this.resetDOMElement();
            }
        };
        HQCard.css = {
            cardSize: DV.config.cssClass.cardSizeStandard,
            cardType: DV.config.cssClass.cardHQ,
            cardWrapper: DV.config.cssClass.cardWrapperStandard,
            scriptTagSelector: 'script[type="x-dv-card-heroicquest"]'
        };
        SQCard.prototype.resetDOMElement =
            HQCard.prototype.resetDOMElement = function(){
                const D = dom;
                if(!this.e){
                    this.e = D.div();
                }else{
                    D.clearElement(this.e);
                    this.e.className = '';
                }
                const cf = DV.config.cssClass;
                D.addClass(this.e, cf.dvCard, this.constructor.css.cardSize,
                           this.constructor.css.cardType);
                D.append(this.e,D.addClass(D.div(), 'card-id'));
                D.append(this.e,
                         D.addClass(D.section(), 'card-name'),
                         D.addClass(D.section(), 'card-body'),
                         D.addClass(D.section(), 'card-qp'));
                return this;
            };
        SQCard.newWrapper = function(){
            return dom.addClass(dom.div(), SQCard.css.cardWrapper);
        };
        HQCard.newWrapper = function(){
            return dom.addClass(dom.div(), HQCard.css.cardWrapper);
        };

        SQCard.prototype.populateFromJSON =
            HQCard.prototype.populateFromJSON = function(json){
                this.resetDOMElement();
                const E = (s)=>this.e.querySelector(s), D = dom;
                if(json.isOfficial) D.addClass(this.e, 'dv-official');
                D.append(E('.card-id'), json.id || '??');
                D.append(E('.card-name'), json.name || '??!');
                const cb = simCardBody(json.body),
                      be = E('.card-body'),
                      bodyInner = D.div();
                /* bodyInner: extra level of indirection needed to get
                   vertical centering working using flexbox layout. */
                D.append(be, bodyInner);
                if(cb.text){
                    DV.runDVMarkup(bodyInner, cb.text, false);
                }
                DV.setFontSizeHint(be, cb.fontSize);
                D.append(E('.card-qp'), json.qp || '?');
                /*if(json.banner){
                  D.append(this.e,
                  D.append(
                  D.addClass(D.span(), 'bottom-banner'),
                  json.banner
                  )
                  );
                  }*/
                this.json = json;
                return this;
            };
    })(global.DarkVenture);
})(window);
